module BundledWebResources

import Artifacts
import Dates
import Downloads
import HTTP
import MIMEs
import PackageExtensionCompat
import Preferences
import SHA
import Scratch

export LocalResource
export Resource
export ResourceRouter
export @comptime

# These are the only currently supported platforms for `bun`. We don't have
# artifacts other platforms so we need to make sure to include the artifact"bun"
# macro only when on the right platforms otherwise the package would not load.
@static if Sys.isapple() || Sys.islinux()
    function bun(; kws...)
        path = joinpath(Artifacts.artifact"bun", "bun$(Sys.iswindows() ? ".exe" : "")")
        return Cmd(Cmd([path]); env = copy(ENV), kws...) # Somewhat replicating JLLWrapper behavior.
    end
else
    function bun(; kws...)
        error("unsupported platform")
    end
end

function default_on_change()
    @debug "bun build finished"
end

# Wrapper for better printing and user interaction with the watcher object.
struct Watcher
    f::Function
end

function Base.show(io::IO, w::Watcher)
    running = istaskstarted(w.f.shutdown.task) && !istaskdone(w.f.shutdown.task)
    print(io, "$(Watcher)(running = $running)")
end

Base.close(w::Watcher) = w.f()

"""
    watch(;
        root::AbstractString = pwd(),
        entrypoint::AbstractString = "input.ts",
        outdir::AbstractString = "dist",
        after_rebuild::Function,
    ) -> Watcher

Runs `bun build --watch` in the given `root` directory, watching the
`entrypoint` file and writing to the `outdir` directory. `after_rebuild` is
called whenever the build finishes.  This is a zero-argument function that can
be used to run any code after the rebuild finishes, such as browser
auto-reloaders.

The returned `Watcher` object can be `close`d to stop the watcher.
"""
function watch(;
    root::AbstractString = pwd(),
    entrypoint::AbstractString = "index.ts",
    outdir::AbstractString = "dist",
    after_rebuild::Function = default_on_change,
)
    if isdir(root)
        cd(root) do
            if isfile(entrypoint)
                if !isdir(outdir)
                    @info "'$(outdir)' directory does not exist. Creating..."
                    mkpath(outdir)
                end

                bin = bun(; dir = pwd())

                inp = Base.PipeEndpoint()
                out = Base.PipeEndpoint()
                err = Base.PipeEndpoint()

                process = run(
                    `$(bin) build $(entrypoint) --outdir $(outdir) --watch`,
                    inp,
                    out,
                    err;
                    wait = false,
                )

                closed = Ref(false)
                task = @async begin
                    @info "running `bun build` watcher"
                    while !closed[] && process_running(process)
                        msg = strip(String(readavailable(out)))
                        if !isempty(msg)
                            rebuilt = false
                            for each in readdir(outdir)
                                if contains(msg, each)
                                    rebuilt = true
                                    break
                                end
                            end
                            if rebuilt
                                after_rebuild()
                            else
                                printstyled("\n$msg\n"; color = :red)
                            end
                        end
                    end
                    @info "`bun build` watcher exited"
                end

                function shutdown()
                    closed[] = true
                    @info "closing `bun build` watcher"
                    kill(process)
                    process_running(process) || Base.close(process)
                    Base.close(inp)
                    Base.close(out)
                    Base.close(err)
                    return wait(task)
                end

                # Ensure graceful shutdown on exit to avoid orphaned `bun`s.
                atexit() do
                    if !closed[]
                        shutdown()
                    end
                end

                return Watcher() do
                    if closed[]
                        error("`bun build` watcher already closed.")
                    else
                        shutdown()
                    end
                end
            else
                error("'$entrypoint' is not a file.")
            end
        end
    else
        error("'$root' is not a directory.")
    end
end

abstract type AbstractResource end
struct LocalResource{T<:AbstractString} <: AbstractResource
    root::T
    path::String
end

content(r::LocalResource) = read(joinpath(r.root, r.path), String)

function Base.pathof(r::LocalResource)
    path = join(splitpath(r.path), '/')
    content = read(joinpath(r.root, r.path))
    hash = string(Base.hash(content); base = 62)
    return "/$path?v=$hash"
end

"""
    @comptime ex

Evaluate an expression at compile time. This is useful for
constructing `Resource` objects at compile time, e.g.:

```julia
my_resource() = @comptime Resource("https://example.com/my_resource.txt"; sha256="...")
```

while still allowing the value to be "`Revise`-able", since it isn't a global
constant, and instead is a function return value.
"""
macro comptime(ex)
    Core.eval(__module__, ex)
end

"""
    Resource(url::String; name::String, sha256::String)

A remote resource that is downloaded and cached locally.
"""
struct Resource <: AbstractResource
    name::String
    url::String
    content::String
    hash::String

    function Resource(url::String; name::String = basename(url), sha256::String = "")
        computed_sha256 = _download_and_cache(url, sha256)
        if sha256 == computed_sha256
            content = read(joinpath(_SCRATCHSPACE[], sha256), String)
            return new(name, url, content, sha256)
        else
            error(
                "SHA256 mismatch for $(repr(url)): expected $(repr(sha256)), got $(repr(computed_sha256)).",
            )
        end
    end
end

content(r::Resource) = r.content

Base.show(io::IO, r::Resource) = print(io, "Resource($(repr(r.url)), $(repr(r.hash)))")

function Base.pathof(r::Resource)
    path = join(("resource", string(hash(r.hash); base = 62), r.name), "/")
    return "/$path"
end

"""
    ResourceRouter(mod::Module)
    
A middleware that serves resources defined in `mod` at the paths returned by
`pathof` for each resource.
"""
ResourceRouter(mod::Module) = _resource_router(mod)

function _resource_router(mod)
    map = _resource_map(mod)
    return function (handler)
        return function (req::HTTP.Request)
            return _resource_request_handler(map, handler, req)
        end
    end
end

function _resource_request_handler(map, handler, req)
    uri = HTTP.URIs.URI(req.target)
    path = uri.path
    if req.method == "GET" && haskey(map, path)
        content_type, body = map[path]
        return HTTP.Response(200, ["Content-Type" => content_type], body)
    else
        return handler(req)
    end
end

function _resource_map(mod::Module)
    map = Dict{String,Tuple{String,String}}()
    for name in names(mod; all = true)
        if isdefined(mod, name) && !Base.isdeprecated(mod, name)
            object = getfield(mod, name)
            if isa(object, Function)
                T = Core.Compiler.return_type(object, Tuple{})
                if T <: AbstractResource && T !== Union{}
                    resource = object()
                    path = pathof(resource)
                    uri = HTTP.URIs.URI(path)
                    mime = MIMEs.mime_from_path(uri.path)
                    content_type = MIMEs.contenttype_from_mime(mime)
                    map[uri.path] = (content_type, content(resource))
                end
            end
        end
    end
    return map
end

function _download_and_cache(url::String, sha256::String)
    cached_path = joinpath(_SCRATCHSPACE[], sha256)
    if isfile(cached_path)
        return sha256
    else
        buffer = IOBuffer()
        Downloads.download(url, buffer)
        seekstart(buffer)
        hash = bytes2hex(SHA.sha256(buffer))
        path = joinpath(_SCRATCHSPACE[], hash)
        seekstart(buffer)
        write(path, buffer)
        return hash
    end
end

function _gc_cached_downloads!()
    last_gc_file = joinpath(_SCRATCHSPACE[], "last_gc")
    if isfile(last_gc_file)
        last_gc_str = read(last_gc_file, String)
        if isempty(last_gc_str)
            write(last_gc_file, string(Dates.now()))
        else
            last_gc = tryparse(Dates.DateTime, last_gc_str)
            last_gc isa Dates.DateTime ||
                error("Invalid 'last_gc' file: $last_gc_str. Must be a valid DateTime.")

            gc_interval_unit = Preferences.@load_preference("gc_interval_unit", "months")

            units = Dict(
                "years" => Dates.Year,
                "months" => Dates.Month,
                "weeks" => Dates.Week,
                "days" => Dates.Day,
                "hours" => Dates.Hour,
                "minutes" => Dates.Minute,
                "seconds" => Dates.Second,
            )

            haskey(units, gc_interval_unit) || error(
                "Invalid 'gc_interval_unit': $gc_interval_unit. Must be one of $(keys(units)).",
            )

            gc_interval_int = Preferences.@load_preference("gc_interval", 1)
            gc_interval_int isa Integer ||
                error("Invalid 'gc_interval': $gc_interval_int. Must be an integer.")

            gc_interval_int < 1 &&
                error("Invalid 'gc_interval': $gc_interval_int. Must be greater than 0.")

            gc_interval_period = units[gc_interval_unit](gc_interval_int)

            if (last_gc + gc_interval_period) < Dates.now()
                @info "Garbage collecting cached downloads. You can adjust the frequency of this operation by setting the 'gc_interval' ($(gc_interval_int)) and 'gc_interval_unit' ($(gc_interval_unit)) preferences."
                files = readdir(_SCRATCHSPACE[]; join = true)
                if length(files) > 1
                    for file in readdir(_SCRATCHSPACE[]; join = true)
                        rm(joinpath(_SCRATCHSPACE[], file))
                    end
                    @info "Garbage collection complete. $(length(files) - 1) files removed."
                    write(last_gc_file, string(Dates.now()))
                else
                    @info "No files to garbage collect."
                end
            end
        end
    else
        write(last_gc_file, string(Dates.now()))
    end
end

const _SCRATCHSPACE = Ref{String}(Scratch.@get_scratch!("download_cache"))

function __init__()
    PackageExtensionCompat.@require_extensions
    _SCRATCHSPACE[] = Scratch.@get_scratch!("download_cache")
    isdir(_SCRATCHSPACE[]) || mkdir(_SCRATCHSPACE[])
    _gc_cached_downloads!()
    return nothing
end

end # module WebResourceCaching
